`include "defines.vh"
module control_unit #(
        parameter INSTR_WIDTH  = 16,   // instructions are 16 bits in width
        parameter DATA_WIDTH   = 8,    // registers are 8 bits in width
        parameter I_ADDR_WIDTH = 10,   // 2*1024 bytes of flash (or ROM in our case)
        parameter ADDR_WIDTH   = 16,   // 64KB address space
        parameter D_ADDR_WIDTH = 7,    // 128 bytes of SRAM
        parameter R_ADDR_WIDTH = 5     // 32 registers
    )(
        output reg  [I_ADDR_WIDTH-1:0] program_counter,
        input  wire  [INSTR_WIDTH-1:0] instruction,
        output wire              [4:0] state,
        output wire                    error,
        input  wire                    clk,
        input  wire                    reset,
        output wire [R_ADDR_WIDTH-1:0] rr_addr,
        output wire [R_ADDR_WIDTH-1:0] rd_addr,
        inout  wire   [DATA_WIDTH-1:0] rr_data,
        inout  wire   [DATA_WIDTH-1:0] rd_data,
        output wire                    rr_cs,
        output wire                    rd_cs,
        output wire                    rr_we,
        output wire                    rd_we,
        output wire                    rr_oe,
        output wire                    rd_oe,
        output wire                    alu_enable,
        output reg               [7:0] alu_opsel,
        output wire   [DATA_WIDTH-1:0] alu_flags_in,
        input  wire   [DATA_WIDTH-1:0] alu_flags_out,
        output reg    [DATA_WIDTH-1:0] alu_rr,
        output reg    [DATA_WIDTH-1:0] alu_rd,
        input  wire   [DATA_WIDTH-1:0] alu_out,
        output wire   [ADDR_WIDTH-1:0] bus_addr,
        inout  wire   [DATA_WIDTH-1:0] bus_data,
        output wire                    mem_cs,
        output wire                    mem_we,
        output wire                    mem_oe,
        output wire                    io_cs,
        output wire                    io_we,
        output wire                    io_oe
    );

    wire              [8:0] signals;
    wire              [7:0] opcode_type;
    wire             [16:0] opcode_group;
    wire [R_ADDR_WIDTH-1:0] opcode_rd;
    wire [R_ADDR_WIDTH-1:0] opcode_rr;
    wire             [11:0] opcode_imd;
    wire              [2:0] opcode_bit;
    reg   [INSTR_WIDTH-1:0] instr_buffer;
    reg    [DATA_WIDTH-1:0] alu_out_buffer;
    reg    [DATA_WIDTH-1:0] writeback_value;
    wire   [ADDR_WIDTH-1:0] indirect_addr;
    wire   [DATA_WIDTH-1:0] data_to_store;
    reg    [DATA_WIDTH-1:0] sreg;
    reg    [DATA_WIDTH-1:0] sp;
    reg    [ADDR_WIDTH-1:0] saved_pc;
    wire                    cycle_count;

    state_machine fsm (
        .state(state),
        .cycle_count(cycle_count),
        .opcode_group(opcode_group),
        .clk(clk),
        .reset(reset)
    );

    decode_unit #(
        .INSTR_WIDTH(INSTR_WIDTH)
    ) decode (
        .instruction(instr_buffer),
        .opcode_type(opcode_type),
        .opcode_group(opcode_group),
        .opcode_imd(opcode_imd),
        .opcode_rd(opcode_rd),
        .opcode_rr(opcode_rr),
        .opcode_bit(opcode_bit)
    );

    signal_generation_unit sig (
        .state(state),
        .cycle_count(cycle_count),
        .opcode_type(opcode_type),
        .opcode_group(opcode_group),
        .signals(signals)
    );

    reg_file_interface_unit #(
        .DATA_WIDTH(DATA_WIDTH),
        .INSTR_WIDTH(INSTR_WIDTH),
        .D_ADDR_WIDTH(D_ADDR_WIDTH),
        .R_ADDR_WIDTH(R_ADDR_WIDTH)
    ) rf_int (
        .opcode_type(opcode_type),
        .opcode_rd(opcode_rd),
        .opcode_rr(opcode_rr),
        .signals(signals),
        .writeback_value(writeback_value),
        .rr_addr(rr_addr),
        .rd_addr(rd_addr),
        .rr_data(rr_data),
        .rd_data(rd_data),
        .rr_cs(rr_cs),
        .rd_cs(rd_cs),
        .rr_we(rr_we),
        .rd_we(rd_we),
        .rr_oe(rr_oe),
        .rd_oe(rd_oe)
    );

    bus_interface_unit #(
        .MEM_START_ADDR(8'h40),
        .MEM_STOP_ADDR(8'hBF),
        .IO_START_ADDR(8'h00),
        .IO_STOP_ADDR(8'h3F),
        .DATA_WIDTH(DATA_WIDTH),
        .ADDR_WIDTH(ADDR_WIDTH)
    ) bus_int (
        .opcode_group(opcode_group),
        .opcode_imd(opcode_imd),
        .signals(signals),
        .bus_addr(bus_addr),
        .bus_data(bus_data),
        .mem_cs(mem_cs),
        .mem_we(mem_we),
        .mem_oe(mem_oe),
        .io_cs(io_cs),
        .io_we(io_we),
        .io_oe(io_oe),
        .indirect_addr(indirect_addr),
        .data_to_store(data_to_store)
    );

    assign indirect_addr =
        (opcode_group[`GROUP_LOAD_INDIRECT] ||
         opcode_group[`GROUP_STORE_INDIRECT]) ?
            (opcode_group[`GROUP_STACK] ?
                {8'b0, sp} :
        // else, indirect to memory => X or Y or Z
                {alu_rr, alu_rd}) :
        // else, not indirect
            {ADDR_WIDTH{1'bx}};

    assign data_to_store = 
            signals[`CONTROL_MEM_WRITE] ?
                opcode_type == `TYPE_RCALL ?
                    cycle_count ? saved_pc[7:0]  :
                                  saved_pc[15:8] :
                alu_rr :
            signals[`CONTROL_IO_WRITE] ?
                opcode_group[`GROUP_STACK] ? sp :
                opcode_group[`GROUP_ALU]   ? sreg :
                (opcode_type == `TYPE_OUT) ? alu_rr :
                {DATA_WIDTH{1'bx}} :
            {DATA_WIDTH{1'bx}};
    /* TODO 5: in cazul instruciunilor cbi si sbi trebuie stocata valoarea intoarsa de alu.
	 (tot sub semnalul CONTROL_IO_WRITE)*/

    assign alu_flags_in = sreg;
    /* Bloc de atribuire al sreg-ului */
    always @(posedge clk)
        if (reset)
            sreg <= 8'b0;
        else if (signals[`CONTROL_IO_READ] && opcode_group == `GROUP_ALU)
            sreg <= bus_data;
        else if (alu_enable) begin
            sreg <= alu_flags_out;
        end

    /* Bloc de atribuire al sp-ului */
    always @(posedge clk) begin
        if (reset)
            sp <= 0;
        else if (signals[`CONTROL_IO_READ] && opcode_group[`GROUP_STACK])
            sp <= bus_data;
        else
            if (signals[`CONTROL_POSTDEC])
                sp <= sp - 1;
            else if (signals[`CONTROL_PREINC])
                sp <= sp + 1;
    end

    /* Bloc de atribuire al program counter-ului */
    always @(posedge clk) begin
        if (reset) begin
            program_counter <= 0;
        end else if (state == `STATE_EX) begin
            case (opcode_type)
            `TYPE_RJMP: begin
                program_counter <= program_counter + opcode_imd[9:0];
            end
            `TYPE_BRBS: begin
                if (sreg[opcode_bit] == 1)
                    program_counter <= program_counter + opcode_imd[9:0];
            end
            `TYPE_BRBC: begin
                if (sreg[opcode_bit] == 0)
                    program_counter <= program_counter + opcode_imd[9:0];
            end
            `TYPE_RCALL: begin
                program_counter <= program_counter + opcode_imd[11:0];
            end
            endcase
        end else if (state == `STATE_WB) begin
            case (opcode_type)
            `TYPE_RET: begin
                program_counter <= saved_pc;
            end 
            default: begin
                program_counter <= program_counter + 10'b1;
            end
            endcase
        end
    end

    always @(posedge clk) begin
        if (reset)
            saved_pc <= 0;
        else if (state == `STATE_ID &&
                 opcode_type == `TYPE_RCALL)
            saved_pc <= {6'b0, program_counter} + 12'b1;
        else if (state == `STATE_MEM &&
                 opcode_type == `TYPE_RET)
            if (cycle_count == 0) begin
                saved_pc[7:0]  <= bus_data;
            end else begin
                saved_pc[15:8] <= bus_data;
            end
    end

    always @(posedge clk) begin
        if (reset)
            writeback_value <= {DATA_WIDTH{1'b0}};
        else if (opcode_type == `TYPE_LDI)
            writeback_value <= opcode_imd[DATA_WIDTH-1:0];
        else if (signals[`CONTROL_IO_READ] ||
                 signals[`CONTROL_MEM_READ])
            writeback_value <= bus_data;
        else if (opcode_group[`GROUP_ALU])
            writeback_value <= alu_out_buffer;
        else if (opcode_type == `TYPE_MOV)
            writeback_value <= alu_rr;
    end

    /* Buffer pentru instructiunea citita */
    always @(posedge clk)
        if (state == `STATE_IF)
            instr_buffer <= instruction;

    /* Buffer pentru output-ul UAL-ului */
    always @(posedge clk)
        if (state == `STATE_EX)
            alu_out_buffer <= alu_out;

    /* Buffer pentru rd_data si rr_data */
    always @(posedge clk)
        if (state == `STATE_ID) begin
            /* TODO 5: instructiunile cbi si sbi trebuie sa lucreze cu alu, 
             * deci trebuie sa trimiteti cei doi operanzi dupa modelul de mai jos. 
				 Un operand va fi o masca pe biti.
				 Celalalt va fi valoarea ce trebuie modificata (a carui bit este (re)setat)
				 */
				alu_rd <= rd_data;
				alu_rr <= rr_data;
		  end

    /* TODO 5: semnalul alu_enable trebuie sa fie activ si in cazul instructiunilor cbi si sbi. */
    assign alu_enable = (state == `STATE_EX) &&
                        (opcode_group[`GROUP_ALU]);
    
    /* Set alu_opsel to appropriate operation,
     * according to opcode_type and alu_enable */
    always @* begin
        if (alu_enable == 0)
            alu_opsel = 8'bx;
        else begin
            case (opcode_type)
            /* TODO 5: setati alu_opsel pentru a efectua o operatie AND in cazul instructiunii cbi
             * si o operatie OR in cazul instrucitiunii sbi*/
            `TYPE_ADD:
                alu_opsel = `OPSEL_ADD;
            `TYPE_ADC:
                alu_opsel = `OPSEL_ADC;
            `TYPE_SUB:
                alu_opsel = `OPSEL_SUB;
            `TYPE_AND:
                alu_opsel = `OPSEL_AND;
            `TYPE_EOR:
                alu_opsel = `OPSEL_XOR;
            `TYPE_OR:
                alu_opsel = `OPSEL_OR;
            `TYPE_NEG:
                alu_opsel = `OPSEL_NEG;
            default:
                alu_opsel = `OPSEL_NONE;
            endcase
        end
    end

    assign error = (opcode_type == `TYPE_UNKNOWN);
    
endmodule
